from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_q", "0008_auto_20160224_1026"),
    ]

    operations = [
        migrations.AlterField(
            model_name="schedule",
            name="repeats",
            field=models.IntegerField(
                default=-1,
                help_text="n = n times, -1 = forever",
                verbose_name="Repeats",
            ),
        ),
    ]



import importlib

from django.db.models.signals import post_save
from django.dispatch import Signal, receiver
from django.utils.translation import gettext_lazy as _

from django_q.conf import logger
from django_q.models import Task


@receiver(post_save, sender=Task)
def call_hook(sender, instance, **kwargs):
    if instance.hook:
        f = instance.hook
        if not callable(f):
            try:
                module, func = f.rsplit(".", 1)
                m = importlib.import_module(module)
                f = getattr(m, func)
            except (ValueError, ImportError, AttributeError):
                logger.error(
                    _("malformed return hook '%(hook)s' for [%(name)s]")
                    % {"hook": instance.hook, "name": instance.name}
                )
                return
        try:
            f(instance)
        except Exception as e:
            logger.error(
                _("return hook %(hook)s failed on [%(name)s] because %(error)s")
                % {"hook": instance.hook, "name": instance.name, "error": str(e)}
            )


# args: proc_name
post_spawn = Signal()

# args: task
pre_enqueue = Signal()

# args: func, task
pre_execute = Signal()

# args: task
post_execute = Signal()



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_q", "0002_auto_20150630_1624"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="failure",
            options={
                "verbose_name_plural": "Failed tasks",
                "verbose_name": "Failed task",
            },
        ),
        migrations.AlterModelOptions(
            name="schedule",
            options={
                "verbose_name_plural": "Scheduled tasks",
                "ordering": ["next_run"],
                "verbose_name": "Scheduled task",
            },
        ),
        migrations.AlterModelOptions(
            name="success",
            options={
                "verbose_name_plural": "Successful tasks",
                "verbose_name": "Successful task",
            },
        ),
        migrations.RemoveField(
            model_name="task",
            name="id",
        ),
        migrations.AddField(
            model_name="task",
            name="id",
            field=models.CharField(
                max_length=32, primary_key=True, editable=False, serialize=False
            ),
        ),
    ]



import copy

from boto3 import Session
from botocore.client import ClientError

from django_q.brokers import Broker
from django_q.conf import Conf

QUEUE_DOES_NOT_EXIST = "AWS.SimpleQueueService.NonExistentQueue"


class Sqs(Broker):
    def __init__(self, list_key: str = None):
        self.sqs = None
        super(Sqs, self).__init__(list_key)
        self.queue = self.get_queue()

    def __setstate__(self, state):
        super(Sqs, self).__setstate__(state)
        self.sqs = None
        self.queue = self.get_queue()

    def enqueue(self, task):
        response = self.queue.send_message(MessageBody=task)
        return response.get("MessageId")

    def dequeue(self):
        # sqs supports max 10 messages in bulk
        if Conf.BULK > 10:
            Conf.BULK = 10

        params = {"MaxNumberOfMessages": Conf.BULK, "VisibilityTimeout": Conf.RETRY}

        # sqs long polling
        sqs_config = Conf.SQS
        if "receive_message_wait_time_seconds" in sqs_config:
            wait_time_second = sqs_config.get("receive_message_wait_time_seconds", 20)

            # validation of parameter
            if not isinstance(wait_time_second, int):
                raise ValueError("receive_message_wait_time_seconds should be int")
            if wait_time_second > 20:
                raise ValueError(
                    "receive_message_wait_time_seconds is invalid. Reason: Must be >= 0"
                    " and <= 20"
                )
            params.update({"WaitTimeSeconds": wait_time_second})

        tasks = self.queue.receive_messages(**params)
        if tasks:
            return [(t.receipt_handle, t.body) for t in tasks]

    def acknowledge(self, task_id):
        return self.delete(task_id)

    def queue_size(self) -> int:
        return int(self.queue.attributes["ApproximateNumberOfMessages"])

    def lock_size(self) -> int:
        return int(self.queue.attributes["ApproximateNumberOfMessagesNotVisible"])

    def delete(self, task_id):
        message = self.sqs.Message(self.queue.url, task_id)
        message.delete()

    def fail(self, task_id):
        self.delete(task_id)

    def delete_queue(self):
        self.queue.delete()

    def purge_queue(self):
        self.queue.purge()

    def ping(self) -> bool:
        return "sqs" in self.connection.get_available_resources()

    def info(self) -> str:
        return "AWS SQS"

    @staticmethod
    def get_connection(list_key: str = None) -> Session:
        config_cloned = copy.deepcopy(Conf.SQS)
        if "aws_region" in config_cloned:
            config_cloned["region_name"] = config_cloned["aws_region"]
            del config_cloned["aws_region"]

        if "receive_message_wait_time_seconds" in config_cloned:
            del config_cloned["receive_message_wait_time_seconds"]

        return Session(**config_cloned)

    def get_queue(self):
        self.sqs = self.connection.resource("sqs")

        try:
            # try to return an existing queue by name. If the queue does not
            # exist try to create it.
            return self.sqs.get_queue_by_name(QueueName=self.list_key)
        except ClientError as exp:
            if exp.response["Error"]["Code"] != QUEUE_DOES_NOT_EXIST:
                raise exp

        return self.sqs.create_queue(QueueName=self.list_key)



# Generated by Django 4.1.2 on 2022-11-10 01:35

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_q", "0014_schedule_cluster"),
    ]

    operations = [
        migrations.AlterField(
            model_name="schedule",
            name="schedule_type",
            field=models.CharField(
                choices=[
                    ("O", "Once"),
                    ("I", "Minutes"),
                    ("H", "Hourly"),
                    ("D", "Daily"),
                    ("W", "Weekly"),
                    ("BW", "Biweekly"),
                    ("M", "Monthly"),
                    ("BM", "Bimonthly"),
                    ("Q", "Quarterly"),
                    ("Y", "Yearly"),
                    ("C", "Cron"),
                ],
                default="O",
                max_length=2,
                verbose_name="Schedule Type",
            ),
        ),
    ]



import signal

from django.utils.translation import gettext_lazy as _

from django_q.conf import logger

from .exceptions import TimeoutException


class TimeoutHandler:
    def __init__(self, timeout: int):
        self._timeout = timeout

    def raise_timeout_exception(self, signum, frame):
        raise TimeoutException(
            f"Task exceeded maximum timeout value ({self._timeout} seconds)"
        )

    def __enter__(self):
        # if the timeout is -1, then there is no timeout and the task will always keep running until it's done or manually killed
        if self._timeout == -1:
            return
        try:
            signal.signal(signal.SIGALRM, self.raise_timeout_exception)
            signal.alarm(self._timeout)
        except (
            ValueError,
            AttributeError,
        ):  # AttributeError or ValueError might be raised for Windows users
            logger.debug(_("SIGALARM is not available on your platform"))

    def __exit__(self, exc_type, exc_value, traceback):
        if self._timeout == -1:
            return
        """When getting out of the timeout, reset the alarm, so it won't trigger"""
        try:
            signal.alarm(0)
            signal.signal(signal.SIGALRM, signal.SIG_DFL)
        except (
            ValueError,
            AttributeError,
        ):  # AttributeError or ValueError might be raised for Windows users
            logger.debug(_("SIGALARM is not available on your platform"))



# Standard
import os
import signal
import socket
import uuid
from multiprocessing import Event, Process, Value, current_process
from time import sleep

# Django
from django import core, db
from django.apps.registry import apps

try:
    apps.check_apps_ready()
except core.exceptions.AppRegistryNotReady:
    import django

    django.setup()

from django.utils import timezone
from django.utils.translation import gettext_lazy as _

# Local
from django_q.brokers import Broker, get_broker
from django_q.conf import (
    Conf,
    get_ppid,
    logger,
    prometheus_multiprocess,
    psutil,
    setproctitle,
)
from django_q.humanhash import humanize
from django_q.monitor import monitor
from django_q.pusher import pusher
from django_q.queues import Queue
from django_q.scheduler import scheduler
from django_q.status import Stat, Status
from django_q.worker import worker


class Cluster:
    def __init__(self, broker: Broker = None):
        # Cluster do not need an init or default broker except for testing,
        # The sentinel will create a broker for cluster and utilize ALT_CLUSTERS config in Conf.
        self.broker = broker  # DON'T USE get_broker() to set a default broker here.
        self.sentinel = None
        self.stop_event = None
        self.start_event = None
        self.pid = current_process().pid
        self.cluster_id = uuid.uuid4()
        self.host = socket.gethostname()
        self.timeout = None
        signal.signal(signal.SIGTERM, self.sig_handler)
        signal.signal(signal.SIGINT, self.sig_handler)

    def start(self) -> int:
        if setproctitle:
            setproctitle.setproctitle(f"qcluster {current_process().name} {self.name}")
        # Start Sentinel
        self.stop_event = Event()
        self.start_event = Event()
        self.sentinel = Process(
            target=Sentinel,
            name=f"Process-{uuid.uuid4().hex}",
            args=(
                self.stop_event,
                self.start_event,
                self.cluster_id,
                self.broker,
                self.timeout,
            ),
        )
        self.sentinel.start()
        logger.info(_("Q Cluster %(name)s starting.") % {"name": self.name})
        while not self.start_event.is_set():
            sleep(0.1)
        return self.pid

    def stop(self) -> bool:
        if not self.sentinel.is_alive():
            return False
        logger.info(_("Q Cluster %(name)s stopping.") % {"name": self.name})
        self.stop_event.set()
        self.sentinel.join()
        logger.info(_("Q Cluster %(name)s has stopped.") % {"name": self.name})
        self.start_event = None
        self.stop_event = None
        return True

    def sig_handler(self, signum, frame):
        logger.debug(
            _("%(name)s got signal %(signal)s")
            % {
                "name": current_process().name,
                "signal": Conf.SIGNAL_NAMES.get(signum, "UNKNOWN"),
            }
        )
        self.stop()

    @property
    def stat(self) -> Status:
        if self.sentinel:
            return Stat.get(pid=self.pid, cluster_id=self.cluster_id)
        return Status(pid=self.pid, cluster_id=self.cluster_id)

    @property
    def name(self) -> str:
        return humanize(self.cluster_id.hex)

    @property
    def is_starting(self) -> bool:
        return self.stop_event and self.start_event and not self.start_event.is_set()

    @property
    def is_running(self) -> bool:
        return self.stop_event and self.start_event and self.start_event.is_set()

    @property
    def is_stopping(self) -> bool:
        return (
            self.stop_event
            and self.start_event
            and self.start_event.is_set()
            and self.stop_event.is_set()
        )

    @property
    def has_stopped(self) -> bool:
        return self.start_event is None and self.stop_event is None and self.sentinel


class Sentinel:
    def __init__(
        self,
        stop_event,
        start_event,
        cluster_id,
        broker=None,
        timeout=None,
        start=True,
    ):
        # Make sure we catch signals for the pool
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        self.pid = current_process().pid
        self.cluster_id = cluster_id
        self.parent_pid = get_ppid()
        self.name = current_process().name
        self.broker = broker or get_broker()
        self.reincarnations = 0
        self.tob = timezone.now()
        self.stop_event = stop_event
        self.start_event = start_event
        self.pool_size = Conf.WORKERS
        self.pool = []
        self.timeout = timeout or Conf.TIMEOUT
        self.task_queue = (
            Queue(maxsize=Conf.QUEUE_LIMIT) if Conf.QUEUE_LIMIT else Queue()
        )
        self.result_queue = Queue()
        self.event_out = Event()
        self.monitor = None
        self.pusher = None
        if start:
            self.start()

    def queue_name(self):
        # multi-queue: cluster name is (broker's) queue_name
        return self.broker.list_key if self.broker else "--"

    def start(self):
        self.broker.ping()
        self.spawn_cluster()
        self.guard()

    def status(self) -> str:
        if not self.start_event.is_set() and not self.stop_event.is_set():
            return Conf.STARTING
        elif self.start_event.is_set() and not self.stop_event.is_set():
            if self.result_queue.empty() and self.task_queue.empty():
                return Conf.IDLE
            return Conf.WORKING
        elif self.stop_event.is_set() and self.start_event.is_set():
            if self.monitor.is_alive() or self.pusher.is_alive() or len(self.pool) > 0:
                return Conf.STOPPING
            return Conf.STOPPED

    def spawn_process(self, target, *args) -> Process:
        """
        :type target: function or class
        """
        p = Process(target=target, args=args, name=f"Process-{uuid.uuid4().hex}")
        p.daemon = True
        if target == worker:
            p.daemon = Conf.DAEMONIZE_WORKERS
            p.timer = args[2]
            self.pool.append(p)
        p.start()
        return p

    def spawn_pusher(self) -> Process:
        return self.spawn_process(pusher, self.task_queue, self.event_out, self.broker)

    def spawn_worker(self):
        self.spawn_process(
            worker, self.task_queue, self.result_queue, Value("f", -1), self.timeout
        )

    def spawn_monitor(self) -> Process:
        return self.spawn_process(monitor, self.result_queue, self.broker)

    def reincarnate(self, process):
        """
        :param process: the process to reincarnate
        :type process: Process or None
        """
        # close connections before spawning new process
        if not Conf.SYNC:
            db.connections.close_all()
        if process == self.monitor:
            self.monitor = self.spawn_monitor()
            logger.critical(
                _("reincarnated monitor %(name)s after sudden death")
                % {"name": process.name}
            )
        elif process == self.pusher:
            self.pusher = self.spawn_pusher()
            logger.critical(
                _("reincarnated pusher %(name)s after sudden death")
                % {"name": process.name}
            )
        else:
            # check if prometheus is proper configurated
            prometheus_path = os.getenv(
                "PROMETHEUS_MULTIPROC_DIR", os.getenv("prometheus_multiproc_dir")
            )

            if prometheus_multiprocess and prometheus_path:
                prometheus_multiprocess.mark_process_dead(process.pid)

            self.pool.remove(process)
            self.spawn_worker()
            if process.timer.value == 0:
                # only need to terminate on timeout, otherwise we risk destabilizing
                # the queues
                task_name = ""
                if psutil:
                    try:
                        process_name = psutil.Process(process.pid).name()
                        name_splits = process_name.split(" ")
                        task_name = (
                            name_splits[3]
                            if len(name_splits) >= 4 and name_splits[2] == "processing"
                            else ""
                        )
                    except psutil.NoSuchProcess:
                        pass
                process.terminate()
                if task_name:
                    msg = _(
                        "reincarnated worker %(name)s after timeout while processing task %(task_name)s"
                    ) % {"name": process.name, "task_name": task_name}
                else:
                    msg = _("reincarnated worker %(name)s after timeout") % {
                        "name": process.name
                    }
                logger.critical(msg)
            elif int(process.timer.value) == -2:
                logger.info(_("recycled worker %(name)s") % {"name": process.name})
            else:
                logger.critical(
                    _("reincarnated worker %(name)s after death")
                    % {"name": process.name}
                )

        self.reincarnations += 1

    def spawn_cluster(self):
        self.pool = []
        Stat(self).save()
        # close connections before spawning new process
        if not Conf.SYNC:
            db.connections.close_all()
        # spawn worker pool
        for __ in range(self.pool_size):
            self.spawn_worker()
        # spawn auxiliary
        self.monitor = self.spawn_monitor()
        self.pusher = self.spawn_pusher()
        # set worker cpu affinity if needed
        if psutil and Conf.CPU_AFFINITY:
            set_cpu_affinity(Conf.CPU_AFFINITY, [w.pid for w in self.pool])

    def guard(self):
        logger.info(
            _("%(name)s guarding cluster %(cluster_name)s")
            % {
                "name": current_process().name,
                "cluster_name": humanize(self.cluster_id.hex)
                + f" [{self.queue_name()}]",
            }
        )
        self.start_event.set()
        Stat(self).save()
        logger.info(
            _("Q Cluster %(cluster_name)s running.")
            % {
                "cluster_name": humanize(self.cluster_id.hex)
                + f" [{self.queue_name()}]"
            }
        )
        counter = 0
        cycle = Conf.GUARD_CYCLE  # guard loop sleep in seconds
        # Guard loop. Runs at least once
        while not self.stop_event.is_set() or not counter:
            # Check Workers
            for p in self.pool:
                with p.timer.get_lock():
                    # Are you alive?
                    if not p.is_alive() or p.timer.value == 0:
                        self.reincarnate(p)
                        continue
                    # Decrement timer if work is being done
                    if p.timer.value > 0:
                        p.timer.value -= cycle
            # Check Monitor
            if not self.monitor.is_alive():
                self.reincarnate(self.monitor)
            # Check Pusher
            if not self.pusher.is_alive():
                self.reincarnate(self.pusher)
            # Call scheduler once a minute (or so)
            counter += cycle
            if counter >= 30 and Conf.SCHEDULER:
                counter = 0
                scheduler(broker=self.broker)
            # Save current status
            Stat(self).save()
            sleep(cycle)
        self.stop()

    def stop(self):
        Stat(self).save()
        name = current_process().name
        logger.info(_("%(name)s stopping cluster processes") % {"name": name})
        # Stopping pusher
        self.event_out.set()
        # Wait for it to stop
        while self.pusher.is_alive():
            sleep(0.1)
            Stat(self).save()
        # Put poison pills in the queue
        for __ in range(len(self.pool)):
            self.task_queue.put("STOP")
        self.task_queue.close()
        # wait for the task queue to empty
        self.task_queue.join_thread()
        # Wait for all the workers to exit
        while len(self.pool):
            for p in self.pool:
                if not p.is_alive():
                    self.pool.remove(p)
            sleep(0.1)
            Stat(self).save()
        # Finally stop the monitor
        self.result_queue.put("STOP")
        self.result_queue.close()
        # Wait for the result queue to empty
        self.result_queue.join_thread()
        logger.info(_("%(name)s waiting for the monitor.") % {"name": name})
        # Wait for everything to close or time out
        count = 0
        if not self.timeout:
            self.timeout = 30
        while self.status() == Conf.STOPPING and count < self.timeout * 10:
            sleep(0.1)
            Stat(self).save()
            count += 1
        # Final status
        Stat(self).save()


def set_cpu_affinity(n: int, process_ids: list, actual: bool = not Conf.TESTING):
    """
    Sets the cpu affinity for the supplied processes.
    Requires the optional psutil module.
    :param int n: affinity
    :param list process_ids: a list of pids
    :param bool actual: Test workaround for Travis not supporting cpu affinity
    """
    # check if we have the psutil module
    if not psutil:
        logger.warning(_("Skipping cpu affinity because psutil was not found."))
        return
    # check if the platform supports cpu_affinity
    if actual and not hasattr(psutil.Process(process_ids[0]), "cpu_affinity"):
        logger.warning(
            _("Faking cpu affinity because it is not supported on this platform")
        )
        actual = False
    # get the available processors
    cpu_list = list(range(psutil.cpu_count()))
    # affinities of 0 or gte cpu_count, equals to no affinity
    if not n or n >= len(cpu_list):
        return
    # spread the workers over the available processors.
    index = 0
    for pid in process_ids:
        affinity = []
        for k in range(n):
            if index == len(cpu_list):
                index = 0
            affinity.append(cpu_list[index])
            index += 1
        if psutil.pid_exists(pid):
            p = psutil.Process(pid)
            if actual:
                p.cpu_affinity(affinity)
            logger.info(
                _("%(pid)s will use cpu %(affinity)s")
                % {"pid": pid, "affinity": affinity}
            )



import os
from datetime import datetime, timedelta
from multiprocessing import Event, Value
from unittest import mock

import django
import pytest
from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.test import override_settings
from django.utils import timezone
from django.utils.timezone import is_naive

from django_q.brokers import Broker, get_broker
from django_q.conf import Conf
from django_q.monitor import monitor
from django_q.pusher import pusher
from django_q.queues import Queue
from django_q.scheduler import scheduler
from django_q.tasks import Schedule, fetch
from django_q.tasks import schedule as create_schedule
from django_q.tests.settings import BASE_DIR
from django_q.tests.testing_utilities.multiple_database_routers import (
    TestingMultipleAppsDatabaseRouter,
    TestingReplicaDatabaseRouter,
)
from django_q.utils import add_months, localtime
from django_q.worker import worker

if django.VERSION < (4, 0):
    # pytz is the default in django 3.2. Remove when no support for 3.2
    from pytz import timezone as ZoneInfo
else:
    try:
        from zoneinfo import ZoneInfo
    except ImportError:
        from backports.zoneinfo import ZoneInfo


@pytest.fixture
def broker(monkeypatch) -> Broker:
    """
    Patches the Conf object setting the DJANGO_REDIS attribute allowing a default
    redis configuration.
    """
    monkeypatch.setattr(Conf, "DJANGO_REDIS", "default")
    return get_broker()


@pytest.fixture
def orm_broker(monkeypatch) -> None:
    """Patches the Conf object setting the ORM attribute to a database named default."""
    monkeypatch.setattr(Conf, "ORM", "default")


REPLICA_DATABASE_ROUTERS = [
    f"{TestingReplicaDatabaseRouter.__module__}.{TestingReplicaDatabaseRouter.__name__}"
]
REPLICA_DATABASES = {
    "writable": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    },
    "replica": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    },
}

MULTIPLE_APPS_DATABASE_ROUTERS = [
    f"{TestingMultipleAppsDatabaseRouter.__module__}.{TestingMultipleAppsDatabaseRouter.__name__}"  # noqa: E501
]
MULTIPLE_APPS_DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    },
    "admin": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    },
}


@pytest.mark.django_db
def test_scheduler_daylight_saving_time_daily(broker, monkeypatch):
    # Set up a startdate in the Amsterdam timezone (without dst 1 hour ahead). The
    # 28th of March 2021 is the day when sunlight saving starts (at 2 am)

    monkeypatch.setattr(Conf, "TIME_ZONE", "Europe/Amsterdam")
    tz = ZoneInfo("Europe/Amsterdam")
    broker.list_key = "scheduler_test:q"
    # Let's start a schedule at 1 am on the 27th of March. This is in AMS timezone.
    # So, 2021-03-27 00:00:00 when saved (due to TZ being Amsterdam and saved in UTC)
    start_date = datetime(2021, 3, 27, 1, 0, 0)

    # Create schedule with the next run date on the start date. It will move one day
    # forward when we run the scheduler
    schedule = create_schedule(
        "math.copysign",
        1,
        -1,
        name="test math",
        schedule_type=Schedule.DAILY,
        next_run=start_date,
    )

    # Run scheduler so we get the next run date
    scheduler(broker=broker)
    schedule.refresh_from_db()

    # It's now the day after exactly at midnight UTC
    next_run = schedule.next_run
    assert str(next_run) == "2021-03-28 00:00:00+00:00"

    # In the Amsterdam timezone, it's 1 hour over midnight (+01)
    next_run = next_run.astimezone(tz)
    assert str(next_run) == "2021-03-28 01:00:00+01:00"

    # Run scheduler so we get the next run date
    scheduler(broker=broker)
    schedule.refresh_from_db()

    next_run = schedule.next_run

    assert str(next_run) == "2021-03-28 23:00:00+00:00"
    next_run = next_run.astimezone(tz)
    # In the Amsterdam timezone, it's 1 hour over midnight (+02)
    assert str(next_run) == "2021-03-29 01:00:00+02:00"

    # Run scheduler so we get the next run date
    scheduler(broker=broker)
    schedule.refresh_from_db()

    next_run = schedule.next_run

    assert str(next_run) == "2021-03-29 23:00:00+00:00"
    next_run = next_run.astimezone(tz)
    assert str(next_run) == "2021-03-30 01:00:00+02:00"

    # Create second schedule with the next run date on the start date. It will move
    # one day forward when we run the scheduler
    start_date = datetime(2021, 10, 29, 1, 0, 0)
    schedule = create_schedule(
        "django_q.tests.tasks.word_multiply",
        2,
        name="multiply",
        schedule_type=Schedule.DAILY,
        next_run=start_date,
    )

    # Run scheduler so we get the next run date
    scheduler(broker=broker)
    schedule.refresh_from_db()

    next_run = schedule.next_run

    assert str(next_run) == "2021-10-29 23:00:00+00:00"
    # In the Amsterdam timezone, it's 1 hour over midnight (+02)
    next_run = next_run.astimezone(tz)
    assert str(next_run) == "2021-10-30 01:00:00+02:00"

    # Run scheduler so we get the next run date
    scheduler(broker=broker)
    schedule.refresh_from_db()

    next_run = schedule.next_run

    assert str(next_run) == "2021-10-30 23:00:00+00:00"
    # In the Amsterdam timezone, it's 1 hour over midnight (+02)
    next_run = next_run.astimezone(tz)
    assert str(next_run) == "2021-10-31 01:00:00+02:00"

    # Run scheduler so we get the next run date
    scheduler(broker=broker)
    schedule.refresh_from_db()

    next_run = schedule.next_run

    assert str(next_run) == "2021-11-01 00:00:00+00:00"
    # In the Amsterdam timezone, it's 1 hour over midnight (+01)
    # Switch of DST
    next_run = next_run.astimezone(tz)
    assert str(next_run) == "2021-11-01 01:00:00+01:00"


@pytest.mark.django_db
def test_scheduler(broker, monkeypatch):
    broker.list_key = "scheduler_test:q"
    broker.delete_queue()
    schedule = create_schedule(
        "math.copysign",
        1,
        -1,
        name="test math",
        hook="django_q.tests.tasks.result",
        schedule_type=Schedule.HOURLY,
        repeats=1,
    )
    assert schedule.last_run() is None
    # check duplicate constraint
    with pytest.raises(IntegrityError):
        schedule = create_schedule(
            "math.copysign",
            1,
            -1,
            name="test math",
            hook="django_q.tests.tasks.result",
            schedule_type=Schedule.HOURLY,
            repeats=1,
        )
    # run scheduler
    scheduler(broker=broker)
    # set up the workflow
    task_queue = Queue()
    stop_event = Event()
    stop_event.set()
    # push it
    pusher(task_queue, stop_event, broker=broker)
    assert task_queue.qsize() == 1
    assert broker.queue_size() == 0
    task_queue.put("STOP")
    # let a worker handle them
    result_queue = Queue()
    worker(task_queue, result_queue, Value("b", -1))
    assert result_queue.qsize() == 1
    result_queue.put("STOP")
    # store the results
    monitor(result_queue)
    assert result_queue.qsize() == 0
    schedule = Schedule.objects.get(pk=schedule.pk)
    assert schedule.repeats == 0
    assert schedule.last_run() is not None
    assert schedule.success() is True
    assert schedule.next_run < timezone.now() + timedelta(hours=1)
    task = fetch(schedule.task)
    assert task is not None
    assert task.success is True
    assert task.result < 0
    # Once schedule with delete
    once_schedule = create_schedule(
        "django_q.tests.tasks.word_multiply",
        2,
        word="django",
        schedule_type=Schedule.ONCE,
        repeats=-1,
        hook="django_q.tests.tasks.result",
    )
    assert hasattr(once_schedule, "pk") is True
    # negative repeats
    always_schedule = create_schedule(
        "django_q.tests.tasks.word_multiply",
        2,
        word="django",
        schedule_type=Schedule.DAILY,
        repeats=-1,
        hook="django_q.tests.tasks.result",
    )
    assert hasattr(always_schedule, "pk") is True
    # Minute schedule
    minute_schedule = create_schedule(
        "django_q.tests.tasks.word_multiply",
        2,
        word="django",
        schedule_type=Schedule.MINUTES,
        minutes=10,
    )
    assert hasattr(minute_schedule, "pk") is True
    # Cron schedule
    cron_schedule = create_schedule(
        "django_q.tests.tasks.word_multiply",
        2,
        word="django",
        schedule_type=Schedule.CRON,
        cron="0 22 * * 1-5",
    )
    assert hasattr(cron_schedule, "pk") is True
    assert cron_schedule.full_clean() is None
    assert cron_schedule.__str__() == "django_q.tests.tasks.word_multiply"
    with pytest.raises(ValidationError):
        create_schedule(
            "django_q.tests.tasks.word_multiply",
            2,
            word="django",
            schedule_type=Schedule.CRON,
            cron="0 22 * * 1-12",
        )
    # All other types
    for t in Schedule.TYPE:
        if t[0] == Schedule.CRON:
            continue
        schedule = create_schedule(
            "django_q.tests.tasks.word_multiply",
            2,
            word="django",
            schedule_type=t[0],
            repeats=1,
            hook="django_q.tests.tasks.result",
        )
        assert schedule is not None
        assert schedule.last_run() is None
        scheduler(broker=broker)
    # via model
    Schedule.objects.create(
        func="django_q.tests.tasks.word_multiply",
        args="2",
        kwargs='word="django"',
        schedule_type=Schedule.DAILY,
    )
    # scheduler
    scheduler(broker=broker)
    # ONCE schedule should be deleted
    assert Schedule.objects.filter(pk=once_schedule.pk).exists() is False
    # Catch up On
    monkeypatch.setattr(Conf, "CATCH_UP", True)
    now = timezone.now()
    schedule = create_schedule(
        "django_q.tests.tasks.word_multiply",
        2,
        word="catch_up",
        schedule_type=Schedule.HOURLY,
        next_run=timezone.now() - timedelta(hours=12),
        repeats=-1,
    )
    scheduler(broker=broker)
    schedule = Schedule.objects.get(pk=schedule.pk)
    assert schedule.next_run < now
    # Catch up off
    monkeypatch.setattr(Conf, "CATCH_UP", False)
    scheduler(broker=broker)
    schedule = Schedule.objects.get(pk=schedule.pk)
    assert schedule.next_run > now
    # Done
    broker.delete_queue()

    # test bimonthly
    schedule = create_schedule(
        "django_q.tests.tasks.word_multiply",
        2,
        word="catch_up",
        schedule_type=Schedule.BIMONTHLY,
    )
    scheduler(broker=broker)
    schedule = Schedule.objects.get(pk=schedule.pk)
    assert schedule.next_run.date() == add_months(timezone.now(), 2).date()

    # test biweekly
    schedule = create_schedule(
        "django_q.tests.tasks.word_multiply",
        2,
        word="catch_up",
        schedule_type=Schedule.BIWEEKLY,
    )
    scheduler(broker=broker)
    schedule = Schedule.objects.get(pk=schedule.pk)
    assert schedule.next_run.date() == (timezone.now() + timedelta(weeks=2)).date()
    broker.delete_queue()

    monkeypatch.setattr(Conf, "CLUSTER_NAME", "some_cluster_name")
    # create a schedule on another cluster
    schedule = create_schedule(
        "math.copysign",
        1,
        -1,
        name="test schedule on a another cluster",
        hook="django_q.tests.tasks.result",
        schedule_type=Schedule.HOURLY,
        cluster="some_other_cluster_name",
        repeats=1,
    )
    # run scheduler
    scheduler(broker=broker)
    # set up the workflow
    task_queue = Queue()
    stop_event = Event()
    stop_event.set()
    # push it
    pusher(task_queue, stop_event, broker=broker)

    # queue must be empty
    assert task_queue.qsize() == 0

    monkeypatch.setattr(Conf, "CLUSTER_NAME", "default")
    # create a schedule on the same cluster
    schedule = create_schedule(
        "math.copysign",
        1,
        -1,
        name="test schedule with no cluster",
        hook="django_q.tests.tasks.result",
        schedule_type=Schedule.HOURLY,
        cluster="default",
        repeats=1,
    )
    # run scheduler
    scheduler(broker=broker)
    # set up the workflow
    task_queue = Queue()
    stop_event = Event()
    stop_event.set()
    # push it
    pusher(task_queue, stop_event, broker=broker)

    # queue must contain a task
    assert task_queue.qsize() == 1


@pytest.mark.django_db
def test_intended_schedule_kwarg(broker, monkeypatch):
    broker.list_key = "scheduler_test:q"
    broker.delete_queue()
    run_date = timezone.now() - timedelta(hours=1)
    schedule = create_schedule(
        "math.copysign",
        1,
        -1,
        name="test math",
        hook="django_q.tests.tasks.result",
        schedule_type=Schedule.HOURLY,
        repeats=1,
        next_run=run_date,
        intended_date_kwarg="intended_date",
    )
    assert schedule.last_run() is None
    assert schedule.intended_date_kwarg == "intended_date"
    # run scheduler
    scheduler(broker=broker)
    # set up the workflow
    task_queue = Queue()
    stop_event = Event()
    stop_event.set()
    # push it
    pusher(task_queue, stop_event, broker=broker)
    assert task_queue.qsize() == 1
    task = task_queue.get()
    assert "intended_date" in task["kwargs"]
    assert task["kwargs"]["intended_date"] == run_date.isoformat()


@override_settings(
    DATABASE_ROUTERS=REPLICA_DATABASE_ROUTERS, DATABASES=REPLICA_DATABASES
)
@pytest.mark.django_db
def test_scheduler_atomic_must_specify_the_write_db(
    orm_broker: Broker,
):
    """
    GIVEN a environment with a read/write configured replica database
    WHEN the scheduler is called
    THEN the transaction must be called with the write database.
    """
    broker = get_broker(list_key="scheduler_test:q")
    with mock.patch("django_q.cluster.db.transaction") as mocked_db:
        scheduler(broker=broker)
        mocked_db.atomic.assert_called_with(using="writable")


@override_settings(
    DATABASE_ROUTERS=MULTIPLE_APPS_DATABASE_ROUTERS, DATABASES=MULTIPLE_APPS_DATABASES
)
@pytest.mark.django_db
def test_scheduler_atomic_must_specify_the_database_based_on_router_redirection(
    orm_broker: Broker,
):
    """
    GIVEN a environment without a read replica database
    WHEN the scheduler is called
    THEN the transaction atomic must be called using the default connection.
    """
    broker = get_broker(list_key="scheduler_test:q")
    with mock.patch("django_q.cluster.db.transaction") as mocked_db:
        scheduler(broker=broker)
        mocked_db.atomic.assert_called_with(using="default")


@pytest.mark.django_db
def test_schedule_save_sets_next_run_for_cron():
    """Ensure Schedule.save() sets next_run correctly for CRON schedules."""
    cron_expression = "0 12 * * *"  # Executes dialy at 12pm
    schedule = Schedule(
        func="math.sqrt",
        schedule_type=Schedule.CRON,
        cron=cron_expression,
    )

    assert schedule.next_run is not None
    assert schedule.pk is None

    initial_next_run = schedule.next_run
    schedule.save()

    # After save, next_run must be recalculated based on the CRON expression
    assert schedule.next_run > initial_next_run


@pytest.mark.django_db
def test_schedule_save_direct_db(broker):
    """Ensure Schedule.save() updates next_run correctly when created directly in DB."""
    cron_expression = "0 12 * * *"  # Executes dialy at 12pm

    # Creating schedule directly in database
    schedule = Schedule.objects.create(
        func="math.sqrt",
        schedule_type=Schedule.CRON,
        cron=cron_expression,
    )

    # Next run must be defined after execution time
    assert schedule.next_run is not None
    assert schedule.next_run > timezone.now()

    scheduler(broker)

    assert broker.queue_size() == 0


def test_localtime():
    assert not is_naive(localtime())


@override_settings(USE_TZ=False)
def test_naive_localtime():
    assert is_naive(localtime())



import picklefield.fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("django_q", "0009_auto_20171009_0915"),
    ]

    operations = [
        migrations.AlterField(
            model_name="task",
            name="args",
            field=picklefield.fields.PickledObjectField(
                editable=False, null=True, protocol=-1
            ),
        ),
        migrations.AlterField(
            model_name="task",
            name="kwargs",
            field=picklefield.fields.PickledObjectField(
                editable=False, null=True, protocol=-1
            ),
        ),
        migrations.AlterField(
            model_name="task",
            name="result",
            field=picklefield.fields.PickledObjectField(
                editable=False, null=True, protocol=-1
            ),
        ),
    ]



import django.utils.timezone
import picklefield.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Schedule",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        auto_created=True,
                        serialize=False,
                        primary_key=True,
                    ),
                ),
                (
                    "func",
                    models.CharField(
                        max_length=256, help_text="e.g. module.tasks.function"
                    ),
                ),
                (
                    "hook",
                    models.CharField(
                        null=True,
                        blank=True,
                        max_length=256,
                        help_text="e.g. module.tasks.result_function",
                    ),
                ),
                (
                    "args",
                    models.CharField(
                        null=True,
                        blank=True,
                        max_length=256,
                        help_text="e.g. 1, 2, 'John'",
                    ),
                ),
                (
                    "kwargs",
                    models.CharField(
                        null=True,
                        blank=True,
                        max_length=256,
                        help_text="e.g. x=1, y=2, name='John'",
                    ),
                ),
                (
                    "schedule_type",
                    models.CharField(
                        verbose_name="Schedule Type",
                        choices=[
                            ("O", "Once"),
                            ("H", "Hourly"),
                            ("D", "Daily"),
                            ("W", "Weekly"),
                            ("M", "Monthly"),
                            ("Q", "Quarterly"),
                            ("Y", "Yearly"),
                        ],
                        default="O",
                        max_length=1,
                    ),
                ),
                (
                    "repeats",
                    models.SmallIntegerField(
                        verbose_name="Repeats",
                        default=-1,
                        help_text="n = n times, -1 = forever",
                    ),
                ),
                (
                    "next_run",
                    models.DateTimeField(
                        verbose_name="Next Run",
                        default=django.utils.timezone.now,
                        null=True,
                    ),
                ),
                ("task", models.CharField(editable=False, null=True, max_length=100)),
            ],
            options={
                "verbose_name": "Scheduled task",
                "ordering": ["next_run"],
            },
        ),
        migrations.CreateModel(
            name="Task",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        auto_created=True,
                        serialize=False,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(editable=False, max_length=100)),
                ("func", models.CharField(max_length=256)),
                ("hook", models.CharField(null=True, max_length=256)),
                (
                    "args",
                    picklefield.fields.PickledObjectField(editable=False, null=True),
                ),
                (
                    "kwargs",
                    picklefield.fields.PickledObjectField(editable=False, null=True),
                ),
                (
                    "result",
                    picklefield.fields.PickledObjectField(editable=False, null=True),
                ),
                ("started", models.DateTimeField(editable=False)),
                ("stopped", models.DateTimeField(editable=False)),
                ("success", models.BooleanField(editable=False, default=True)),
            ],
        ),
        migrations.CreateModel(
            name="Failure",
            fields=[],
            options={
                "verbose_name": "Failed task",
                "proxy": True,
            },
            bases=("django_q.task",),
        ),
        migrations.CreateModel(
            name="Success",
            fields=[],
            options={
                "verbose_name": "Successful task",
                "proxy": True,
            },
            bases=("django_q.task",),
        ),
    ]
